/*

Conversation widget for kdm greeter

Copyright (C) 2008 Dirk Mueller <mueller@kde.org>
Copyright (C) 2008 Oswald Buddenhagen <ossi@kde.org>

based on classic kdm greeter:

  Copyright (C) 1997, 1998, 2000 Steffen Hansen <hansen@kde.org>
  Copyright (C) 2000-2003 Oswald Buddenhagen <ossi@kde.org>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

*/

#include "kgreet_generic.h"

#include <kglobal.h>
#include <klocale.h>
#include <klineedit.h>
#include <kuser.h>

#include <QLayout>
#include <QLabel>
#include <QTextDocument>

extern KDE_EXPORT KGreeterPluginInfo kgreeterplugin_info; // defined at bottom

static int echoMode;

KGenericGreeter::KGenericGreeter(KGreeterPluginHandler *_handler,
                                 QWidget *parent,
                                 const QString &_fixedEntity,
                                 Function _func, Context _ctx) :
    QObject(),
    KGreeterPlugin(_handler),
    m_label( 0 ),
    m_messageLabel( 0 ),
    m_lineEdit(0),
    fixedUser(_fixedEntity),
    func(_func),
    ctx(_ctx),
    exp(-1),
    running(false)
{
    QGridLayout *grid = 0;
    if (!_handler->gplugHasNode( "pam-prompt" ) ||
            !_handler->gplugHasNode( "user-pw-entry" ) ||
            !_handler->gplugHasNode( "pam-message" )) {
        parent = new QWidget( parent );
        parent->setObjectName( "talker" );
        widgetList << parent;
        grid = new QGridLayout( parent );
        grid->setMargin( 0 );
    }

    if (ctx == ExUnlock || ctx == ExChangeTok)
        fixedUser = KUser().loginName();

    m_label = new QLabel( parent );
    if( grid ) {
        m_label->setMinimumWidth(
                m_label->fontMetrics().width( "A username" ) );
        grid->addWidget( m_label, 0, 0 );
    } else {
        m_label->setObjectName( "pam-prompt" );
        widgetList << m_label;

        m_lineEdit = new KLineEdit( parent );
        m_label->setBuddy( m_lineEdit );
        if( grid ) {
            m_lineEdit->setMinimumWidth(
                    m_lineEdit->fontMetrics().width( "This is a long password" ) );
            grid->addWidget( m_lineEdit, 0, 1 );
        } else {
            m_lineEdit->setObjectName( "user-pw-entry" );
            widgetList << m_lineEdit;
        }
        m_messageLabel = new QLabel( parent );
        if( grid ) {
            m_messageLabel->setMinimumWidth(
                    m_messageLabel->fontMetrics().width( "Information text" ) );
            grid->addWidget( m_messageLabel, 1, 0 );
        } else {
            m_messageLabel->setObjectName( "pam-message" );
            widgetList << m_messageLabel;
        }
    }
}

// virtual
KGenericGreeter::~KGenericGreeter()
{
    abort();
    qDeleteAll( widgetList );
}

void // virtual
KGenericGreeter::loadUsers(const QStringList &users)
{
    m_users = users;
}

void // virtual
KGenericGreeter::presetEntity(const QString &entity, int /* field */)
{
    // assert(!running);
    curUser = entity;
}

QString // virtual
KGenericGreeter::getEntity() const
{
    return fixedUser.isEmpty() ? curUser : fixedUser;
}

void // virtual
KGenericGreeter::setUser(const QString &user)
{
    // assert(running);
    // assert(fixedUser.isEmpty());
    if (!(kgreeterplugin_info.flags & KGreeterPluginInfo::Presettable))
        return; // Not interested in PAM telling us who logged in
    if (exp) {
        abort();
        start();
    }
    curUser = user;
    if (m_lineEdit) { // could be null if plugin is misconfigured
        m_lineEdit->setText(user);
        m_lineEdit->selectAll();
        m_lineEdit->setFocus();
    }
}

void // virtual
KGenericGreeter::setEnabled(bool enable)
{
    // assert(func == Authenticate && ctx == Shutdown);
    // XXX this is likely to bear some bogosity
    m_lineEdit->setEnabled( enable );
    if (enable && m_lineEdit)
        m_lineEdit->setFocus();
}

bool // virtual
KGenericGreeter::textMessage(const char *text, bool err)
{
    if (err)
        return false;

    if (m_infoMsgs.isEmpty())
        revive();
    QString qtext = QString::fromUtf8(text);
    m_infoMsgs.append(qtext);
    m_messageLabel->setText( m_infoMsgs.join( "\n" ));

    return true;
}

void // virtual
KGenericGreeter::textPrompt(const char *prompt, bool echo, bool /* nonBlocking */)
{
    exp =
        exp >= 0 ||
        func != Authenticate ||
		!echo ||
        !(kgreeterplugin_info.flags & KGreeterPluginInfo::Presettable);

    if (!exp && !fixedUser.isEmpty()) {
        handler->gplugReturnText(fixedUser.toLocal8Bit(),
                                 KGreeterPluginHandler::IsUser);
        return;
    }

    if (m_infoMsgs.isEmpty())
        revive();
    else
        m_infoMsgs.clear();
    m_children.append(m_label);
    m_echo = echo;

    m_label->setText( QString::fromUtf8( prompt ).trimmed());
    m_lineEdit->setContextMenuPolicy( Qt::NoContextMenu );
    m_lineEdit->setCompletionObject( 0 );
    m_lineEdit->disconnect( this );
    m_lineEdit->clear();
    m_lineEdit->setPasswordMode( false );
    m_lineEdit->setEchoMode( QLineEdit::Normal );
    if (echo) {
        if (!exp) {
            if (m_label->text().isEmpty())
                m_label->setText( i18n("&Username:"));
            if (!m_users.isEmpty()) {
                KCompletion *userNamesCompletion = new KCompletion;
                userNamesCompletion->setItems(m_users);
                m_lineEdit->setCompletionObject(userNamesCompletion);
                m_lineEdit->setAutoDeleteCompletionObject(true);
                m_lineEdit->setCompletionMode(KGlobalSettings::CompletionAuto);
            }
            if (!curUser.isEmpty()) {
                m_lineEdit->setText(curUser);
                m_lineEdit->selectAll();
            }
            connect(m_lineEdit, SIGNAL(editingFinished()), SLOT(slotLoginLostFocus()));
        }
        connect(m_lineEdit, SIGNAL(editingFinished()), SLOT(slotChanged()));
        connect(m_lineEdit, SIGNAL(textChanged(QString)), SLOT(slotChanged()));
    } else {
        if (::echoMode == -1)
            m_lineEdit->setPasswordMode( true );
        else
            m_lineEdit->setEchoMode( ::echoMode ? QLineEdit::Password : QLineEdit::NoEcho );
        connect( m_lineEdit, SIGNAL(editingFinished()), SLOT(slotChanged()) );
    }
    m_lineEdit->setFocus();
}

bool // virtual
KGenericGreeter::binaryPrompt(const char *, bool)
{
    // FIXME
    return true;
}

void // virtual
KGenericGreeter::start()
{
    exp = -1;
    running = true;
    handler->gplugStart();
}

void // virtual
KGenericGreeter::suspend()
{
}

void // virtual
KGenericGreeter::resume()
{
}

void // virtual
KGenericGreeter::next()
{
    if (m_lineEdit) {
        m_lineEdit->setEnabled(false);
        QString text = m_lineEdit->text();
        handler->gplugReturnText(text.toLocal8Bit(),
                                 !m_echo ?
                                     KGreeterPluginHandler::IsSecret :
                                     !exp ?
                                         KGreeterPluginHandler::IsUser : 0);
    }
}

void // virtual
KGenericGreeter::abort()
{
    running = false;
    if (exp >= 0) {
        exp = -1;
        handler->gplugReturnText(0, 0);
    }
}

void // virtual
KGenericGreeter::succeeded()
{
    failed(); // redefining terms :-D
}

void // virtual
KGenericGreeter::failed()
{
    // assert(running || timed_login);
    if (!m_infoMsgs.isEmpty()) {
        QString text = "<qt>";
        foreach (const QString &msg, m_infoMsgs)
            text += "<p>" + Qt::escape(msg) + "</p>";
        text += "</qt>";
        revive();
        handler->gplugMsgBox(QMessageBox::Information, text);
    } else {
        m_lineEdit->setEnabled( false );
    }
    exp = -1;
    running = false;
}

void // virtual
KGenericGreeter::revive()
{
    // assert(!running);
    m_infoMsgs.clear();
    m_lineEdit->clear();
    m_lineEdit->setEnabled( true );
}

void // virtual
KGenericGreeter::clear()
{
    // assert(!running && !passwd1Edit);
    revive();
    curUser = QString::null;
}


// private
void
KGenericGreeter::slotLoginLostFocus()
{
    m_lineEdit->setText(m_lineEdit->text().trimmed());
    if (curUser != m_lineEdit->text()) {
        curUser = m_lineEdit->text();
        handler->gplugSetUser(curUser);
    }
}

void
KGenericGreeter::slotChanged()
{
    handler->gplugChanged();
}

// factory
static bool init(const QString &,
                 QVariant(*getConf)(void *, const char *, const QVariant &),
                 void *ctx)
{
    echoMode = getConf(ctx, "EchoMode", QVariant(-1)).toInt();
    // Fielded entities are not supported per se.
    // This implies that the first field is the presettable entity, if any.
    kgreeterplugin_info.flags |= KGreeterPluginInfo::Presettable;
    KGlobal::locale()->insertCatalog("kgreet_generic");
    return true;
}

static void done(void)
{
    KGlobal::locale()->removeCatalog("kgreet_generic");
}

static KGreeterPlugin *
create(KGreeterPluginHandler *handler,
       QWidget *parent,
       const QString &fixedEntity,
       KGreeterPlugin::Function func,
       KGreeterPlugin::Context ctx)
{
    return new KGenericGreeter(handler, parent, fixedEntity, func, ctx);
}

KDE_EXPORT KGreeterPluginInfo kgreeterplugin_info = {
    I18N_NOOP2("@item:inmenu authentication method", "Generic"), "generic",
    KGreeterPluginInfo::Local,
    init, done, create
};

#include "kgreet_generic.moc"
